/*
 * Copyright (C) 2003, 2004, 2005, 2006
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/*----------------------------------------------------------------
 * project ....: LTI Digital Image/Signal Processing Library
 * file .......: ltiFastEigenSystem_template.h
 * authors ....: Peter Doerfler
 * organization: LTI, RWTH Aachen
 * creation ...: 08.05.2003
 * revisions ..: $Id: ltiFastGeneralizedEigenSystem_template.h,v 1.5 2006/09/11 17:03:16 ltilib Exp $
 */


#include "ltiValidator.h"

// do not include this file directly!

#undef _LTI_DEBUG
//#define _LTI_DEBUG 3
#include "ltiDebug.h"

namespace lti
{
  /// default constructor
  template <class T>
  fastGeneralizedEigenSystem<T>::fastGeneralizedEigenSystem()
    : generalizedEigenSystem<T>(), lapackInterface() {
    parameters p;
    setParameters(p);
  }

  /// constructor, sets the parameters
  template <class T>
  fastGeneralizedEigenSystem<T>
  ::fastGeneralizedEigenSystem(const parameters& theParams)
    : generalizedEigenSystem<T>(), lapackInterface() {
    setParameters(theParams);
  }

  /// constructor, sets the parameters
  template <class T>
  fastGeneralizedEigenSystem<T>
  ::fastGeneralizedEigenSystem(const int dimensions)
    : generalizedEigenSystem<T>(), lapackInterface() {
    parameters p;
    p.dimensions=dimensions;
    setParameters(p);
  }

  /// destructor
  template <class T>
  fastGeneralizedEigenSystem<T>::~fastGeneralizedEigenSystem() {
  }


  template <class T>
  functor* fastGeneralizedEigenSystem<T>::clone() const {
    return new fastGeneralizedEigenSystem<T>(*this);
  }


  template <class T>
  const typename fastGeneralizedEigenSystem<T>::parameters&
  fastGeneralizedEigenSystem<T>::getParameters() const {
    const parameters* params =
      dynamic_cast<const parameters*>(&functor::getParameters());

    if (isNull(params)) {
      throw functor::invalidParametersException(getTypeName());
    }

    return *params;
  }


  template <class T>
  bool fastGeneralizedEigenSystem<T>::apply(const matrix<T>& a,
                                            const matrix<T>& b,
                                            vector<T>& eigenvalues) const
  {

    const parameters& param = getParameters();
    bool success;

    lti::validator validate;

    if (!validate.apply(a)) {
      _lti_debug("\n-- Matrix A is invalid! --\n");
      this->setStatusString("Matrix A has invalid elements!\n");
      return false;
    } else {
      _lti_debug("+\n");
    }

    if (!validate.apply(b)) {
      _lti_debug("\n-- Matrix B is invalid! --\n");
      this->setStatusString("Matrix B has invalid elements!\n");
      return false;
    } else {
      _lti_debug("+\n");
    }

    integer itype;
    switch (param.problemType) {
      case parameters::Type1:
        itype=1;
        break;
      case parameters::Type2:
        itype=2;
        break;
      case parameters::Type3:
        itype=3;
        break;
      default:
        this->setStatusString("Illegal value of parameter problemType");
        return false;
    }

    char jobz='N';
    char uplo='U';
    integer n=a.columns();
    integer lda=n;
    integer ldb=b.columns();
    integer il=1;
    integer iu=n;
    integer info=0;


    matrix<T> tmpA;
    tmpA.transpose(a);
    if (tmpA.getMode() != matrix<T>::Connected) {
      throw matrixNotConnected();
    }
    matrix<T> tmpB;
    tmpB.transpose(b);
    if (tmpB.getMode() != matrix<T>::Connected) {
      throw matrixNotConnected();
    }

    // workspace variables
    T twork[1];
    integer lwork=-1;
    T* work=0;
    integer* iwork=0;

    // find all eigenvectors and eigenvalues
    if (param.dimensions==0 || param.dimensions>=n) {
      
      _lti_debug2("get all eigenvalues/-vectors\n");
      
      eigenvalues.resize(n);

      //use divide and conquer
      if (param.useDC) {
        _lti_debug2("divide and conquer\n");
        integer tiwork[1];
        integer liwork=-1;
        sygvd(&itype, &jobz, &uplo, 
              &n, &tmpA.at(0,0), &lda, &tmpB.at(0,0), &ldb, 
              &eigenvalues.at(0), &twork[0], &lwork, 
              &tiwork[0], &liwork, &info);
        lwork=int(twork[0]);
        work = new T[lwork];
        liwork=tiwork[0];
        iwork = new integer[liwork];

        sygvd(&itype, &jobz, &uplo, 
              &n, &tmpA.at(0,0), &lda, &tmpB.at(0,0), &ldb, 
              &eigenvalues.at(0), &work[0], &lwork, &iwork[0], &liwork, &info);

      } else {
        _lti_debug2("no divide and conquer\n");
        sygv(&itype, &jobz, &uplo, 
             &n, &tmpA.at(0,0), &lda, &tmpB.at(0,0), &ldb, 
             &eigenvalues.at(0), &twork[0], &lwork, &info);
        
        lwork=int(twork[0]);
        work = new T[lwork];

        sygv(&itype, &jobz, &uplo, 
             &n, &tmpA.at(0,0), &lda, &tmpB.at(0,0), &ldb, 
             &eigenvalues.at(0), &work[0], &lwork, &info);
      }

   
      
    } else {
      //find some eigenvalues/-vectors
      
      _lti_debug2("find some eigenvalues/-vectors\n");
      T vl=T(0);
      T vu=T(0);
      T abstol=T(0);
      integer m=0;
      il=n-param.dimensions+1;
      char range='I';
      integer ifail=0;
      integer ldz=1;
      T* dummy=new T[1];

      eigenvalues.resize(iu-il+1);
      
      iwork = new integer[5*n];

      sygvx(&itype, &jobz, &range, &uplo, 
            &n, &tmpA.at(0,0), &lda, &tmpB.at(0,0), &ldb, 
            &vl, &vu, &il, &iu, &abstol, &m, 
            &eigenvalues.at(0), &dummy[0], &ldz,
            &twork[0], &lwork, &iwork[0], &ifail, &info);
      
      lwork=int(twork[0]);
      work = new T[lwork];
      
      sygvx(&itype, &jobz, &range, &uplo, 
            &n, &tmpA.at(0,0), &lda, &tmpB.at(0,0), &ldb, 
            &vl, &vu, &il, &iu, &abstol, &m, 
            &eigenvalues.at(0), &dummy[0], &ldz,
            &work[0], &lwork, &iwork[0], &ifail, &info);

      delete[] dummy;
    }

    int tmpRows,i;
    tmpRows = eigenvalues.size();
      
    //swap eigenvalues
    T ttt;
    for(i=0; i<tmpRows/2; i++) {
      ttt=eigenvalues.at(i);
      eigenvalues.at(i)=eigenvalues.at(tmpRows-1-i);
      eigenvalues.at(tmpRows-1-i)=ttt;
    }


    success=(info==0);

    delete[] work;
    delete[] iwork;

    return success;

  }

  template <class T>
  bool fastGeneralizedEigenSystem<T>::apply(const matrix<T>& a,
                                            const matrix<T>& b,
                                            vector<T>& eigenvalues,
                                            matrix<T>& eigenvectors) const
  {

    // do things here
    //std::cout<<"running"<<endl;
    const parameters& param = getParameters();
    bool success;

    //cerr << "S";

    //debug("Computing Eigenvectors of this matrix:\n" << theMatrix);

    lti::validator validate;

    if (!validate.apply(a)) {
      _lti_debug("\n-- Matrix A is invalid! --\n");
      this->setStatusString("Matrix A has invalid elements!\n");
      return false;
    } else {
      _lti_debug("+\n");
    }

    if (!validate.apply(b)) {
      _lti_debug("\n-- Matrix B is invalid! --\n");
      this->setStatusString("Matrix B has invalid elements!\n");
      return false;
    } else {
      _lti_debug("+\n");
    }

    integer itype;
    switch (param.problemType) {
      case parameters::Type1:
        itype=1;
        break;
      case parameters::Type2:
        itype=2;
        break;
      case parameters::Type3:
        itype=3;
        break;
      default:
        this->setStatusString("Illegal value of parameter problemType");
        return false;
    }

    char jobz='V';
    char uplo='U';
    integer n=a.columns();
    integer lda=n;
    integer ldb=b.columns();
    integer il=1;
    integer iu=n;
    integer info=0;


    matrix<T> tmpB;
    tmpB.transpose(b);
    if (tmpB.getMode() != matrix<T>::Connected) {
      throw matrixNotConnected();
    }

    // workspace variables
    T twork[1];
    integer lwork=-1;
    T* work=0;
    integer* iwork=0;

    // find all eigenvectors and eigenvalues
    if (param.dimensions==0 || param.dimensions>=n) {
      
      _lti_debug2("get all eigenvalues/-vectors\n");
      
      eigenvectors.transpose(a);
      eigenvalues.resize(n);

      //use divide and conquer
      if (param.useDC) {
        _lti_debug2("divide and conquer\n");
        integer tiwork[1];
        integer liwork=-1;
        sygvd(&itype, &jobz, &uplo, 
              &n, &eigenvectors.at(0,0), &lda, &tmpB.at(0,0), &ldb, 
              &eigenvalues.at(0), &twork[0], &lwork, 
              &tiwork[0], &liwork, &info);
        lwork=int(twork[0]);
        work = new T[lwork];
        liwork=tiwork[0];
        iwork = new integer[liwork];

        sygvd(&itype, &jobz, &uplo, 
              &n, &eigenvectors.at(0,0), &lda, &tmpB.at(0,0), &ldb, 
              &eigenvalues.at(0), &work[0], &lwork, &iwork[0], &liwork, &info);

      } else {
        _lti_debug2("no divide and conquer\n");
        sygv(&itype, &jobz, &uplo, 
             &n, &eigenvectors.at(0,0), &lda, &tmpB.at(0,0), &ldb, 
             &eigenvalues.at(0), &twork[0], &lwork, &info);
        
        lwork=int(twork[0]);
        work = new T[lwork];

        sygv(&itype, &jobz, &uplo, 
             &n, &eigenvectors.at(0,0), &lda, &tmpB.at(0,0), &ldb, 
             &eigenvalues.at(0), &work[0], &lwork, &info);
      }

   
      
    } else {
      //find some eigenvalues/-vectors
      
      _lti_debug2("find some eigenvalues/-vectors\n");
      T vl=T(0);
      T vu=T(0);
      T abstol=T(0);
      integer m=0;
      il=n-param.dimensions+1;
      char range='I';
      integer ifail=0;
      integer ldz=n;

      matrix<T> tmpA;
      tmpA.transpose(a);
      if (tmpA.getMode() != matrix<T>::Connected) {
        throw matrixNotConnected();
      }
      eigenvectors.resize(iu-il+1,ldz);
      eigenvalues.resize(iu-il+1);
      
      iwork = new integer[5*n];

      sygvx(&itype, &jobz, &range, &uplo, 
            &n, &tmpA.at(0,0), &lda, &tmpB.at(0,0), &ldb, 
            &vl, &vu, &il, &iu, &abstol, &m, 
            &eigenvalues.at(0), &eigenvectors.at(0,0), &ldz,
            &twork[0], &lwork, &iwork[0], &ifail, &info);
      
      lwork=int(twork[0]);
      work = new T[lwork];
      
      sygvx(&itype, &jobz, &range, &uplo, 
            &n, &tmpA.at(0,0), &lda, &tmpB.at(0,0), &ldb, 
            &vl, &vu, &il, &iu, &abstol, &m, 
            &eigenvalues.at(0), &eigenvectors.at(0,0), &n,
            &work[0], &lwork, &iwork[0], &ifail, &info);

    }

    //transpose and sort in descending order
    int tmpRows,tmpCols,i,j;
    tmpRows = eigenvectors.rows();
    tmpCols = eigenvectors.columns();
      
    matrix<T> temp;
      
    if (eigenvectors.getMode() == matrix<T>::Connected) {
      eigenvectors.detach(temp); // transport data to 'temp'
    }
    else {
      temp.copy(eigenvectors);
    }
      
    eigenvectors.resize(tmpCols,tmpRows,T(),false,false);
    for(i=0;i<tmpRows;++i) {
      for(j=0;j<tmpCols;++j) {
        eigenvectors.at(j,i)=temp.at(tmpRows-1-i,j);
      }
    }
      
    //swap eigenvalues
    T ttt;
    for(i=0; i<tmpRows/2; i++) {
      ttt=eigenvalues.at(i);
      eigenvalues.at(i)=eigenvalues.at(tmpRows-1-i);
      eigenvalues.at(tmpRows-1-i)=ttt;
    }


    success=(info==0);

    delete[] work;
    delete[] iwork;

    return success;
  }


  template<>
  int fastGeneralizedEigenSystem<float>::sygv(integer* itype, char* jobz, 
                                              char* uplo, integer* n, 
                                              float* a, integer* lda,
                                              float* b, integer* ldb,
                                              float* w,
                                              float* work, integer* lwork,
                                              integer* info) const;

  template<>
  int fastGeneralizedEigenSystem<double>::sygv(integer* itype, char* jobz, 
                                               char* uplo, integer* n,
                                               double* a, integer* lda, 
                                               double* b, integer* ldb, 
                                               double* w,
                                               double* work, integer* lwork,
                                               integer* info) const;

  template<>
  int fastGeneralizedEigenSystem<float>::sygvd(integer* itype, char* jobz, 
                                               char* uplo, integer* n, 
                                               float* a, integer* lda,
                                               float* b, integer* ldb,
                                               float* w,
                                               float* work, integer* lwork,
                                               integer* iwork, integer* liwork,
                                               integer* info) const;
  
  template<>
  int fastGeneralizedEigenSystem<double>::sygvd(integer* itype, char* jobz, 
                                                char* uplo, integer* n, 
                                                double* a, integer* lda, 
                                                double* b, integer* ldb, 
                                                double* w,
                                                double* work, integer* lwork,
                                                integer* iwork, integer* liwork,
                                                integer* info) const;

  template<>
  int fastGeneralizedEigenSystem<float>::sygvx(integer* itype, char* jobz, 
                                               char* range, char* uplo, 
                                               integer* n, 
                                               float* a, integer* lda,
                                               float* b, integer* ldb,
                                               float* vl, float* vu,
                                               integer* il, integer* iu,
                                               float* abstol,
                                               integer* m, float* w, 
                                               float* z, integer* ldz,
                                               float* work, integer* lwork,
                                               integer* iwork, integer* ifail,
                                               integer* info) const;

  template<>
  int fastGeneralizedEigenSystem<double>::sygvx(integer* itype, char* jobz, 
                                                char* range, char* uplo, 
                                                integer* n, 
                                                double* a, integer* lda,
                                                double* b, integer* ldb,
                                                double* vl, double* vu,
                                                integer* il, integer* iu,
                                                double* abstol,
                                                integer* m, double* w, 
                                                double* z, integer* ldz,
                                                double* work, integer* lwork,
                                                integer* iwork, integer* ifail,
                                                integer* info) const;
  
}

#include "ltiUndebug.h"
