/*
 * Copyright (C) 2004, 2005, 2006
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.
 */


/*--------------------------------------------------------------------
 * project ....: LTI-Lib: Image Processing and Computer Vision Library
 * file .......: ltiBayerDemosaicing.h
 * authors ....: Arnd Hannemann 
 * organization: LTI, RWTH Aachen
 * creation ...: 28.10.2004
 * revisions ..: $Id: ltiBayerDemosaicing.h,v 1.4 2006/12/31 22:29:55 alvarado Exp $
 */

#ifndef _LTI_BAYER_DEMOSAICING_H_
#define _LTI_BAYER_DEMOSAICING_H_



#include "ltiMatrix.h"
#include "ltiVector.h"

#include "ltiTransform.h"

namespace lti {
  /**
   * This functor makes a color interpolation of common
   * BAYER Patterns used in digital cameras.
   */
  class bayerDemosaicing : public transform {
  public:
    
    /**
     * In a Bayer pattern there are three different color filters
     * used in front of the actual pixels of the CCD chip. They have
     * the colors red(R), green(G), and blue(B). On a 2x2 grid of
     * pixels two are always green and these always lie on a
     * diagonal. This results in four possible constellations for
     * the three colors.
     *
     * If NoBayer is chosen then no demosaicing takes place. This
     * makes sense for mono cameras or to avoid demosaicing.
     */
    enum eBayerPattern {
      NoBayer, /**< no Bayer pattern used, no demosaicing, mono camera */
      RGGB,    /**< Red, Green in first row, Green, Blue in second */
      BGGR,    /**< Blue, Green in first row, Green, Red in second */
      GBRG,    /**< Green, Blue in first row, Red, Green in second */
      GRBG     /**< Green, Red in first row, Blue, Green in second */
    };
    
    /**
     * Enumeration of supported interpolation methods 
     */
    enum eBayerMethod {
      Simple,         /**< Do averaging of two green values per pixel */
      NearestNeighbor /**< Take the nearest (vertical) green value for each
                           pixel */
    };


    /**
     * The parameters for the class bayerDemosaicing
     */
    class parameters : public transform::parameters {
    public:


      /**
       * Default constructor
       */
      parameters();

      /**
       * Copy constructor
       * @param other the parameters object to be copied
       */
      parameters(const parameters& other);

      /**
       * Destructor
       */
      ~parameters();

      /**
       * Returns name of this type
       */
      const char* getTypeName() const;

      /**
       * Copy the contents of a parameters object
       * @param other the parameters object to be copied
       * @return a reference to this parameters object
       */
      parameters& copy(const parameters& other);

      /**
       * Copy the contents of a parameters object
       * @param other the parameters object to be copied
       * @return a reference to this parameters object
       */
      parameters& operator=(const parameters& other);


      /**
       * Returns a pointer to a clone of the parameters
       */
      virtual functor::parameters* clone() const;

      /**
       * Write the parameters in the given ioHandler
       * @param handler the ioHandler to be used
       * @param complete if true (the default) the enclosing begin/end will
       *        be also written, otherwise only the data block will be written.
       * @return true if write was successful
       */
      virtual bool write(ioHandler& handler,const bool complete=true) const;

      /**
       * Read the parameters from the given ioHandler
       * @param handler the ioHandler to be used
       * @param complete if true (the default) the enclosing begin/end will
       *        be also written, otherwise only the data block will be written.
       * @return true if write was successful
       */
      virtual bool read(ioHandler& handler,const bool complete=true);

#     ifdef _LTI_MSC_6
      /**
       * This function is required by MSVC only, as a workaround for a
       * very awful bug, which exists since MSVC V.4.0, and still by
       * V.6.0 with all bugfixes (so called "service packs") remains
       * there...  This method is also public due to another bug, so please
       * NEVER EVER call this method directly: use read() instead
       */
      bool readMS(ioHandler& handler,const bool complete=true);

      /**
       * This function is required by MSVC only, as a workaround for a
       * very awful bug, which exists since MSVC V.4.0, and still by
       * V.6.0 with all bugfixes (so called "service packs") remains
       * there...  This method is also public due to another bug, so please
       * NEVER EVER call this method directly: use write() instead
       */
      bool writeMS(ioHandler& handler,const bool complete=true) const;
#     endif

      // ------------------------------------------------
      // the parameters
      // ------------------------------------------------

      /**
       * Method used to interpolate unknown pixel values.
       *
       * Default value: Simple
       */
      eBayerMethod method;

      /**
       * Pattern used in input data.
       *
       * Default value: RGGB
       */
      eBayerPattern bayerPattern;

    };

    /**
     * Default constructor
     */
    bayerDemosaicing();

    /**
     * Construct a functor using the given parameters
     */
    bayerDemosaicing(const parameters& par);

    /**
     * Copy constructor
     * @param other the object to be copied
     */
    bayerDemosaicing(const bayerDemosaicing& other);

    /**
     * Destructor
     */
    virtual ~bayerDemosaicing();

    /**
     * Returns the name of this type ("bayerDemosaicing")
     */
    virtual const char* getTypeName() const;

    /**
     * operates on a copy of the given %parameters.
     * @param src matrix<ubyte> with the source data.
     * @param dest matrix<ubyte> where the result will be left.
     * @return true if apply successful or false otherwise.
     */
    bool apply(const matrix<ubyte>& src,matrix<rgbPixel>& dest) const;


    /**
     * Copy data of "other" functor.
     * @param other the functor to be copied
     * @return a reference to this functor object
     */
    bayerDemosaicing& copy(const bayerDemosaicing& other);

    /**
     * Alias for copy member
     * @param other the functor to be copied
     * @return a reference to this functor object
     */
    bayerDemosaicing& operator=(const bayerDemosaicing& other);

    /**
     * Returns a pointer to a clone of this functor.
     */
    virtual functor* clone() const;

    /**
     * Returns used parameters
     */
    const parameters& getParameters() const;

  protected:
    bool simpleDemosaicingRG(const eBayerPattern pattern,
                             const matrix<ubyte>& src, 
                             matrix<rgbPixel>& dest) const;

    bool nearestNeighborDemosaicingRG(const eBayerPattern pattern,
                                      const matrix<ubyte>& src,
                                      matrix<rgbPixel>& dest) const;
    bool simpleDemosaicingBG(const eBayerPattern pattern,
                             const matrix<ubyte>& src, 
                             matrix<rgbPixel>& dest) const;

    bool nearestNeighborDemosaicingBG(const eBayerPattern pattern,
                                      const matrix<ubyte>& src,
                                      matrix<rgbPixel>& dest) const;
    bool generateGrayImage(const matrix<ubyte>& src,
                           matrix<rgbPixel>& dest) const;
  };

  // global read and write functions for enumerations
  /**
   * read function for bayerDemosaicing::eBayerMethod
   *
   * @ingroup gStorable
   */
  bool read(ioHandler& handler, bayerDemosaicing::eBayerMethod& data);

  /**
   * write function for bayerDemosaicing::eBayerMethod
   *
   * @ingroup gStorable
   */
  bool write(ioHandler& handler, const bayerDemosaicing::eBayerMethod& data);

  /**
   * read function for bayerDemosaicing::eBayerPattern
   *
   * @ingroup gStorable
   */
  bool read(ioHandler& handler, bayerDemosaicing::eBayerPattern& data);

  /**
   * write function for bayerDemosaicing::eBayerPattern
   *
   * @ingroup gStorable
   */
  bool write(ioHandler& handler, const bayerDemosaicing::eBayerPattern& data);


}

#endif
