/*
 * Copyright (C) 2002, 2003, 2004, 2005, 2006
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 * 
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.  
 */ 

 
/*--------------------------------------------------------------------
 * project ....: LTI-Lib: Image Processing and Computer Vision Library
 * file .......: ltiCalibrationBlobFeatures.h
 * authors ....: Dorothee Finck
 * organization: LTI, RWTH Aachen
 * creation ...: 28.11.2002
 * revisions ..: $Id: ltiCalibrationBlobFeatures.h,v 1.7 2006/02/07 18:33:59 ltilib Exp $
 */

#ifndef _LTI_CALIBRATION_BLOB_FEATURES_H_
#define _LTI_CALIBRATION_BLOB_FEATURES_H_

#include "ltiGeometricFeaturesFromMask.h"

namespace lti {

  /**
   * To calibrate the camera optic feature points of an calibration
   * pattern must be extracted. This class extracts the center of
   * gravity of e.g. dark circular control points on a white
   * background. The control points must be placed on a rectangular
   * pattern, wich may be recorded under perspective distortion.
   *
   * The output is a matrix where the feature point in the i-th row
   * and j-th column corresponds to the control point in the i-th row
   * and j-th column of the calibration pattern. Because a very simple
   * algorithm for sorting the feature points is used, the perspective
   * distortion is limited in such a way that either all x-coordinates
   * of the points in one column must be smaller than all
   * x-coordinates of the points in the next column or all
   * y-coordinates of all points in one row respectively.
   */
  class calibrationBlobFeatures : public modifier {
  public:
    /**
     * the parameters for the class calibrationBlobFeatures
     */
    class parameters : public modifier::parameters {
    public:
      /**
       * default constructor
       */
      parameters();

      /**
       * copy constructor
       * @param other the parameters object to be copied
       */
      parameters(const parameters& other);

      /**
       * destructor
       */
      ~parameters();

      /**
       * returns name of this type
       */
      const char* getTypeName() const;

      /**
       * copy the contents of a parameters object
       * @param other the parameters object to be copied
       * @return a reference to this parameters object
       */
      parameters& copy(const parameters& other);
   
      /**
       * copy the contents of a parameters object
       * @param other the parameters object to be copied
       * @return a reference to this parameters object
       */
      parameters& operator=(const parameters& other);


      /**
       * returns a pointer to a clone of the parameters
       */
      virtual functor::parameters* clone() const;

      /**
       * write the parameters in the given ioHandler
       * @param handler the ioHandler to be used
       * @param complete if true (the default) the enclosing begin/end will
       *        be also written, otherwise only the data block will be written.
       * @return true if write was successful
       */
      virtual bool write(ioHandler& handler,const bool complete=true) const;

      /**
       * read the parameters from the given ioHandler
       * @param handler the ioHandler to be used
       * @param complete if true (the default) the enclosing begin/end will
       *        be also written, otherwise only the data block will be written.
       * @return true if write was successful
       */
      virtual bool read(ioHandler& handler,const bool complete=true);

#     ifdef _LTI_MSC_6     
      /**
       * this function is required by MSVC only, as a workaround for a
       * very awful bug, which exists since MSVC V.4.0, and still by
       * V.6.0 with all bugfixes (so called "service packs") remains
       * there...  This method is also public due to another bug, so please
       * NEVER EVER call this method directly: use read() instead
       */
      bool readMS(ioHandler& handler,const bool complete=true);

      /**
       * this function is required by MSVC only, as a workaround for a
       * very awful bug, which exists since MSVC V.4.0, and still by
       * V.6.0 with all bugfixes (so called "service packs") remains
       * there...  This method is also public due to another bug, so please
       * NEVER EVER call this method directly: use write() instead
       */
      bool writeMS(ioHandler& handler,const bool complete=true) const;
#     endif

      // ------------------------------------------------
      // the parameters
      // ------------------------------------------------

      /**
       * All pixels with a grey value below this threshold are considered
       * blob pixels, unless the formed object is too small.
       */
      int blobThreshold;

      /**
       * All blobs smaller than this are considered as noise and discarded.
       */
      int minimumBlobSize;

      /**
       * The number of blobs that are expected along the x- and y-axis. It is
       * assumed that all lines and columns of the calibration pattern contain
       * the same number of markers (blobs).
       */
      point numberOfBlobs;
      
      /**
       * If true, all x-coordinates of the points in one row must be smaller
       * than all x-coordinatens of the points in the next row. If false, all
       * y-coordinates of the points in one column must be smaller than all
       * y-coordinatens of the points in the next column.
       */
      bool xOrdering;
    };

    /**
     * default constructor
     */
    calibrationBlobFeatures();

    /**
     * Construct a functor using the given parameters
     */
    calibrationBlobFeatures(const parameters& par);

    /**
     * copy constructor
     * @param other the object to be copied
     */
    calibrationBlobFeatures(const calibrationBlobFeatures& other);

    /**
     * destructor
     */
    virtual ~calibrationBlobFeatures();

    /**
     * returns the name of this type ("calibrationBlobFeatures")
     */
    virtual const char* getTypeName() const;

    /**
     * Computes a matrix with all calibration features. The image will be
     * converted to grey-values by averaging all 3 channels.
     * Returns true on success and false otherwise. E.g. false is returned if
     * the number of expected calibration features could not be extracted.
     */
    bool apply(const image& src, matrix<point>& dest);

    /**
     * Computes a matrix with all calibration features from a grey
     * value image.
     *
     * @return true on success and false otherwise. E.g. false is
     * returned if the number of expected calibration features could
     * not be extracted.
     */
    bool apply(const channel8& src, matrix<point>& dest);
      
    /**
     * copy data of "other" functor.
     * @param other the functor to be copied
     * @return a reference to this functor object
     */
    calibrationBlobFeatures& copy(const calibrationBlobFeatures& other);

    /**
     * alias for copy member
     * @param other the functor to be copied
     * @return a reference to this functor object
     */
    calibrationBlobFeatures& operator=(const calibrationBlobFeatures& other);

    /**
     * returns a pointer to a clone of this functor.
     */
    virtual functor* clone() const;

    /**
     * returns used parameters
     */
    const parameters& getParameters() const;

  };
}

#endif
