/*
 * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007
 * Department of Electronics, ITCR, Costa Rica
 * 
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.  
 */ 

 
/*--------------------------------------------------------------------
 * project ....: LTI-Lib: Image Processing and Computer Vision Library
 * file .......: ltiSplitImageToYPbPr.h
 * authors ....: Pablo Alvarado
 * organization: Electronics Department, ITCR
 * creation ...: 04.01.2007
 * revisions ..: $Id: ltiSplitImageToYPbPr.h,v 1.3 2007/01/05 22:53:33 alvarado Exp $
 */

#ifndef _LTI_SPLIT_IMAGE_TO_Y_Pb_Pr_H_
#define _LTI_SPLIT_IMAGE_TO_Y_Pb_Pr_H_

#include "ltiSplitImage.h"

namespace lti {
  /**
   * Computes the YPbPr values from a given RGB color representation
   * (rgbPixel).
   *
   * In the literature, technical and scientific, there is often confusion
   * among the color spaces YUV, YCbCr and YPbPr.  Poynton in
   * http://www.poynton.com/notes/colour_and_gamma/ColorFAQ.html explains that
   * YUV is usually never correctly meant, because the color space normally
   * used for component digital video is the YCbCr (ITU-RS601 or CCIR-601).
   * Other devices use the YPbPr, but the "real" YUV is rarely employed.
   *
   * The LTI-Lib provides all three spaces:
   *
   * - YCbCr: lti::splitImageToYCbCr used by IEEE 1394 FireWire cameras
   * - YPbPr: lti::splitImageToYPbPr used by some WebCams
   * - YUV:   lti::splitImageToYUV   did they really meant to use this?
   *
   * This functor splits an RGB images into the color space YPbPr
   * using the conversion equations given by the above mentioned reference:
   *
   * \f[
   * \begin{bmatrix}
   * Y \\
   * Pb \\
   * Pr
   * \end{bmatrix}
   * =
   * \begin{bmatrix}
   *  0.299    &  0.587    &  0.114   \\
   * -0.168736 & -0.331264 &  0.5     \\
   *  0.5      & -0.418688 & -0.081312 
   * \end{bmatrix}
   * \begin{bmatrix}
   * R \\
   * G \\
   * B
   * \end{bmatrix}
   * \f]
   *
   * In the formula above, RGB have been assumed to be in the intervall [0,1].
   * With it, the Y values will have also be in [0,1], while Pb and Pr will lie
   * between [-0.5 and 0.5].  For the ubyte types an offset of 128 is added to
   * the Pb and Pr values, which are previously scaled by 255.
   *
   * @ingroup gColor
   */
  class splitImageToYPbPr : public splitImage {
  public:

    /**
     * default constructor
     */
    splitImageToYPbPr();

    /**
     * copy constructor
     * @param other the object to be copied
     */
    splitImageToYPbPr(const splitImageToYPbPr& other);

    /**
     * destructor
     */
    virtual ~splitImageToYPbPr();

    /**
     * returns the name of this type ("splitImageToYPbPr")
     */
    virtual const char* getTypeName() const;
      
    /**
     * copy data of "other" functor.
     * @param other the functor to be copied
     * @return a reference to this functor object
     */
    splitImageToYPbPr& copy(const splitImageToYPbPr& other);

    /**
     * alias for copy member
     * @param other the functor to be copied
     * @return a reference to this functor object
     */
    splitImageToYPbPr& operator=(const splitImageToYPbPr& other);

    /**
     * returns a pointer to a clone of this functor.
     */
    virtual functor* clone() const;


    /**
     *  split pixel into float values (between 0 and 1)
     */
    inline virtual bool apply(const rgbPixel& pixel,
                              float& c1,
                              float& c2,
                              float& c3) const; 

    /**
     * Split pixel into 8-bit values (between 0 and 255)
     *
     * N.B.: when casting the transformation result to unsigned shorts
     * (8-bit channel) major rounding errors will occur.
     *
     * As a result, the merging operation might produce negative
     * values or values > 1, which are truncated subsequently.
     *
     * When accurate Y, U and V values are required, prefer float values
     */
    inline virtual bool apply(const rgbPixel& pixel,
                              ubyte& c1,
                              ubyte& c2,
                              ubyte& c3) const;  

    /**
     *  split pixel into float channels
     */   
    virtual bool apply(const image& img,
                       channel& c1,
                       channel& c2,
		       channel& c3) const;
    
    /**
     * Split image into 8-bit channels.
     *
     * N.B.: when casting the transformation result to unsigned shorts
     * (8-bit channel) major rounding errors will occur.
     *
     * As a result, the merging operation might produce negative
     * values or values > 1, which are truncated subsequently.  When
     * accurate Y, U and V values are required, prefer float values.
     */
    virtual bool apply(const image& img,
                       channel8& c1,
                       channel8& c2,
		       channel8& c3) const;

  };


  //
  // --------------------------------
  // implementation of inline methods
  // --------------------------------
  //

  // split pixel into float values (between 0 and 1)
  inline bool splitImageToYPbPr::apply(const rgbPixel& pixel,
                                       float& c1,
                                       float& c2,
                                       float& c3) const {
    
    const float red   = static_cast<float>( pixel.getRed() );
    const float green = static_cast<float>( pixel.getGreen() );
    const float blue  = static_cast<float>( pixel.getBlue() );
    
    c1 =  red   *  0.299f/255.0f  +
          green *  0.587f/255.0f  +
          blue  *  0.114f/255.0f; // Y
    c2 =  red   *(-0.168736f/255.0f) +
          green *(-0.331264f/255.0f) +
          blue  *  0.5f/255.0f; // U=Pb
    c3 =  red   *  0.5f/255.0f  +
          green *(-0.418688f/255.0f) +
          blue  *(-0.081312f/255.0f); // V=Pr
    
    return true;
  }

  // split pixel into 8-bit values (between 0 and 255)
  // N.B.: when casting the transformation result to unsigned shorts
  // (8-bit channel) major rounding errors will occur.
  // As a result, the merging operation might
  // produce negative values or values > 1,  which are truncated subsequently.
  // When accurate Y, U and V values are required, prefer float values!
  inline bool splitImageToYPbPr::apply(const rgbPixel& pixel,
                                       ubyte& c1,
                                       ubyte& c2,
                                       ubyte& c3) const {
    const int r = pixel.getRed();
    const int g = pixel.getGreen();
    const int b = pixel.getBlue();

    // use fixed point arithmetic for faster code

    // The following coefficients are tuned to produce the smallest possible
    // error of RGB -> YPbPr -> RGB convertions.  Please DO NOT CHANGE! 
   
    // The error cannot be further reduced as the ubyte precision looses many
    // information when converting RGB->YPbPr.  With the given coefficients
    // 3999903 values of all 2^24 available were perfectly retrieved (31.3%),
    // an the mean deviation error is 0.761 (RGB color space L2 distance).

    // The values are the fixed point arithmetic with 16 bits decimals.
    // The offset 128 is multiplied by 65536, but 32768 is added also
    // to force correct rounding.
    c1 = static_cast<ubyte>(( 19595*r + 38470*g +  7471*b + 32768)>>16);
    c2 = static_cast<ubyte>((-11058*r - 21710*g + 32768*b + 8421376)>>16);
    c3 = static_cast<ubyte>(( 32768*r - 27439*g -  5329*b + 8421376)>>16);

    return true;
  }

  // ---------------------------------------
  // end of implementation of inline methods
  // ---------------------------------------

}

#endif
